/*
    ext2_mkfs.c -- ext2 fs creator
    Copyright (C) 1999 Lennert Buytenhek <buytenh@dsv.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

static const char _ext2_mkfs_c[] = "$Id: ext2_mkfs.c,v 1.2 1999/10/05 19:13:59 buytenh Exp $";

#define USE_EXT2_IS_DATA_BLOCK
#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <uuid/uuid.h>
#include "ext2.h"

#define is_sparse() (1)

void ext2_mkfs_write_main(struct ext2_dev_handle *handle,
			  struct ext2_super_block *sb,
			  struct ext2_group_desc *gd)
{
	int i;
	int numgroups;
	int gdblocks;
	unsigned char *sbbuf;

	sbbuf = (unsigned char *)sb;
	if (sb->s_log_block_size)
	{
		sbbuf = malloc(1024 << sb->s_log_block_size);
		memset(sbbuf, 0, 1024 << sb->s_log_block_size);
		memcpy(sbbuf+1024, sb, 1024);
	}

	numgroups = howmany(sb->s_blocks_count - sb->s_first_data_block, sb->s_blocks_per_group);
	gdblocks = howmany(numgroups * sizeof(struct ext2_group_desc), 1024 << sb->s_log_block_size);

	for (i=0;i<numgroups;i++)
	{
		if (is_sparse())
		{
			int offset;

			offset = sb->s_first_data_block + i * sb->s_blocks_per_group;

			handle->ops->write(handle->cookie, sbbuf, offset, 1);
			handle->ops->write(handle->cookie, gd, offset+1, gdblocks);
		}
	}
}

void ext2_mkfs_write_meta(struct ext2_dev_handle *handle,
			  struct ext2_super_block *sb,
			  struct ext2_group_desc *gd)
{
	int i;
	int numgroups;
	int itsize;
	int gdtsize;
	unsigned char *zero;
	unsigned char *bb;
	unsigned char *ib;

	numgroups = howmany(sb->s_blocks_count - sb->s_first_data_block, sb->s_blocks_per_group);
	itsize = howmany(sizeof(struct ext2_inode) * sb->s_inodes_per_group, (1024 << sb->s_log_block_size));
	gdtsize = howmany(sizeof(struct ext2_group_desc) * numgroups, (1024 << sb->s_log_block_size));

	bb = malloc(1024 << sb->s_log_block_size);
	ib = malloc(1024 << sb->s_log_block_size);
	zero = malloc(1024 << sb->s_log_block_size);

	memset(zero, 0, 1024 << sb->s_log_block_size);

	for (i=0;i<numgroups;i++)
	{
		int admin;
		int groupsize;
		int groupoffset;
		int j;
		int offset;

		admin = itsize + 2;
		if (is_sparse())
			admin += gdtsize + 1;

		groupoffset = i*sb->s_blocks_per_group + sb->s_first_data_block;
		groupsize = min(sb->s_blocks_count - offset, sb->s_blocks_per_group);

		{
			memset(bb, 0, 1024 << sb->s_log_block_size);
			if (is_sparse())
				for (j=0;j<gdtsize+1;j++)
					bb[j>>3] |= _bitmap[j&7];

			for (j=0;j<itsize+2;j++)
			{
				int k = j + gdtsize + 1;

				bb[k>>3] |= _bitmap[k&7];
			}

			for (j=groupsize;j<(1 << (sb->s_log_block_size + 3));j++)
				bb[j>>3] |= _bitmap[j&7];

			handle->ops->write(handle->cookie, bb, groupoffset + gdtsize + 1, 1);

		}

		{
			memset(ib, 0, 1024 << sb->s_log_block_size);
			if (!i)
				for (j=0;j<12;j++)
					ib[j>>3] |= _bitmap[j&7];

			for (j=sb->s_inodes_per_group;j<(1 << (sb->s_log_block_size + 3));j++)
				bb[j>>3] |= _bitmap[j&7];

			handle->ops->write(handle->cookie, ib, groupoffset + gdtsize + 2, 1);
		}

		handle->ops->write(handle->cookie, zero, groupoffset + gdtsize + 3, itsize);

		gd[i].bg_block_bitmap = groupoffset + gdtsize + 1;
		gd[i].bg_inode_bitmap = groupoffset + gdtsize + 2;
		gd[i].bg_inode_table = groupoffset + gdtsize + 3;
		gd[i].bg_free_blocks_count = sb->s_blocks_per_group - admin;
		gd[i].bg_free_inodes_count = sb->s_inodes_per_group;
		gd[i].bg_used_dirs_count = 0;
		gd[i].bg_pad = 0;
		gd[i].bg_reserved[0] = 0;
		gd[i].bg_reserved[1] = 0;
		gd[i].bg_reserved[2] = 0;

		sb->s_free_blocks_count = sb->s_blocks_per_group - admin;
	}
}

void ext2_mkfs_create_root_inode(struct ext2_fs *fs)
{
#if 0
	struct ext2_inode i;

	memset(&i, 0, sizeof(struct ext2_inode));
	i.i_mode = S_IFDIR | 0755;
	i.i_uid = 0;
	i.i_size = fs->blocksize;
	i.i_atime = 0;
	i.i_ctime = 0;
	i.i_mtime = 0;
	i.i_dtime = 0;
	i.i_gid = 0;
	i.i_links_count = 0;
	i.i_blocks = fs->blocksize >> 9;
	i.i_flags = 0;
	i.i_mode = 0;

	ext2_write_inode(fs, 2, &i);
#endif
}

struct ext2_fs *ext2_mkfs(struct ext2_dev_handle *handle,
			  blk_t numblocks,
			  int log_block_size,
			  blk_t blocks_per_group,
			  int inodes_per_group,
			  int sparse_sb,
			  int reserved_block_percentage)
{
	struct ext2_super_block sb;
	struct ext2_group_desc *gd;
	int numgroups;
	int first_block;

	if (!numblocks)
		numblocks = handle->ops->get_size(handle->cookie);

	if (!log_block_size)
		log_block_size = 12;

	if (!blocks_per_group)
		blocks_per_group = 8 * (1 << log_block_size);

	if (!inodes_per_group)
		inodes_per_group = 1024;

	if (!reserved_block_percentage)
		reserved_block_percentage = 5;

	handle->ops->set_blocksize(handle->cookie, log_block_size);
	first_block = (log_block_size==10)?1:0;
	numgroups = howmany(numblocks - first_block, blocks_per_group);
	gd = malloc(numgroups * sizeof(struct ext2_group_desc) + (1 << log_block_size));

	memset(&sb, 0, 1024);

	sb.s_inodes_count = numgroups * inodes_per_group;
	sb.s_blocks_count = numblocks;
	sb.s_r_blocks_count = ((u_int64_t)numblocks * reserved_block_percentage) / 100;
	sb.s_free_blocks_count = 0;
	sb.s_free_inodes_count = sb.s_inodes_count - 1;
	sb.s_first_data_block = first_block;
	sb.s_log_block_size = log_block_size - 10;
	sb.s_log_frag_size = sb.s_log_block_size;
	sb.s_blocks_per_group = blocks_per_group;
	sb.s_frags_per_group = blocks_per_group;
	sb.s_inodes_per_group = inodes_per_group;
	sb.s_mtime = 0;
	sb.s_wtime = 0;
	sb.s_mnt_count = 0;
	sb.s_max_mnt_count = 30;
	sb.s_magic = 0xEF53;
	sb.s_state = EXT2_VALID_FS;
	sb.s_errors = 0;
	sb.s_minor_rev_level = 1;
	sb.s_lastcheck = 0;
	sb.s_checkinterval = 0;
	sb.s_creator_os = 0;
	sb.s_rev_level = 0;
	sb.s_def_resuid = 0;
	sb.s_def_resgid = 0;
	sb.s_first_ino = 12;
	sb.s_inode_size = 128;
	sb.s_block_group_nr = 0;
	sb.s_feature_compat = 0;
	sb.s_feature_incompat = 0;
	sb.s_feature_ro_compat = 0;
	uuid_generate(sb.s_uuid);
	memset(sb.s_volume_name, 0, 16);
	memset(sb.s_last_mounted, 0, 64);
	sb.s_algorithm_usage_bitmap = 0;
	sb.s_prealloc_blocks = 0;
	sb.s_prealloc_dir_blocks = 0;
	sb.s_padding1 = 0;

	ext2_mkfs_write_meta(handle, &sb, gd);
	ext2_mkfs_write_main(handle, &sb, gd);


	{
		struct ext2_fs *fs;

		fs = ext2_open(handle, 0);
		ext2_mkfs_create_root_inode(fs);

		return fs;
	}
}
